package dev.fitko.fitconnect.core.destination;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.jwk.ApiJwk;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.InvalidKeyException;
import dev.fitko.fitconnect.api.services.destination.DestinationService;
import dev.fitko.fitconnect.api.services.destination.KeyService;
import dev.fitko.fitconnect.api.services.portal.SelfServicePortalService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.text.ParseException;
import java.util.UUID;

public class PublicKeyService implements KeyService {

    private static final Logger LOGGER = LoggerFactory.getLogger(PublicKeyService.class);

    private static final ObjectMapper MAPPER = new ObjectMapper();

    private final ApplicationConfig config;
    private final DestinationService destinationService;
    private final ValidationService validationService;

    private final SelfServicePortalService portalService;

    public PublicKeyService(final ApplicationConfig config, final DestinationService destinationService,
                            final SelfServicePortalService portalService, final ValidationService validationService){
        this.config = config;
        this.destinationService = destinationService;
        this.portalService = portalService;
        this.validationService = validationService;
    }

    @Override
    public RSAKey getPublicEncryptionKey(final UUID destinationId) {
        final Destination destination = this.destinationService.getDestination(destinationId);
        final ApiJwk publicKey = this.destinationService.getPublicKey(destinationId, destination.getEncryptionKid());
        final RSAKey rsaKey = toRSAKey(publicKey);
        validateKey(rsaKey);
        return rsaKey;
    }

    @Override
    public RSAKey getPublicSignatureKey(final UUID destinationId, final String keyId) {
        final RSAKey sigKey = toRSAKey(this.destinationService.getPublicKey(destinationId, keyId));
        validateKey(sigKey);
        return sigKey;
    }

    @Override
    public RSAKey getPortalSignatureKey(final String keyId) {
        final RSAKey sigKey = this.portalService.getPublicKeys()
                .getKeys()
                .stream()
                .filter(key -> key.getKid().equals(keyId))
                .map(this::toRSAKey)
                .findFirst()
                .orElse(null);
        validateKey(sigKey);
        return sigKey;
    }

    private void validateKey(final RSAKey rsaKey) {
        final ValidationResult validationResult = this.validationService.validateEncryptionPublicKey(rsaKey);
        if(validationResult.hasError()) {
            if (this.config.getCurrentEnvironment().isAllowInsecurePublicKey()) {
                LOGGER.warn("Invalid public key", validationResult.getError());
            } else {
                throw new InvalidKeyException("Public key is not valid", validationResult.getError());
            }
        }
    }

    private RSAKey toRSAKey(final ApiJwk jwk){
        try {
            return RSAKey.parse(MAPPER.writeValueAsString(jwk));
        } catch (final JsonProcessingException | ParseException e) {
            throw new InvalidKeyException("Key could not be parsed", e);
        }
    }
}
